#pragma once

// solanaceae port of tox_ngc_ft1

#include <solanaceae/toxcore/tox_interface.hpp>
#include <solanaceae/toxcore/tox_event_interface.hpp>

#include "./ngcext.hpp"
#include "./ledbat.hpp"

#include "./rcv_buf.hpp"
#include "./snd_buf.hpp"

#include <map>
#include <set>

// uint32_t - same as tox friend ft
// TODO: fill in toxfriend file kinds
enum class NGCFT1_file_kind : uint32_t {
	//INVALID = 0u, // DATA?

	// id:
	// group (implicit)
	// peer pub key + msg_id
	NGC_HS1_MESSAGE_BY_ID = 1u, // history sync PoC 1
	// TODO: oops, 1 should be avatar v1

	// id: TOX_FILE_ID_LENGTH (32) bytes
	// this is basically and id and probably not a hash, like the tox friend api
	// this id can be unique between 2 peers
	ID = 8u,

	// id: hash of the info, like a torrent infohash (using the same hash as the data)
	// TODO: determain internal format
	// draft: (for single file)
	//   - 256 bytes | filename
	//   - 8bytes | file size
	//   - 4bytes | chunk size
	//   - array of chunk hashes (ids) [
	//     - SHA1 bytes (20)
	//   - ]
	HASH_SHA1_INFO,
	// draft: (for single file) v2
	//   - c-string | filename
	//   - 8bytes | file size
	//   - 4bytes | chunk size
	//   - array of chunk hashes (ids) [
	//     - SHA1 bytes (20)
	//   - ]
	HASH_SHA1_INFO2,
	// draft: multiple files
	//   - 4bytes | number of filenames
	//   - array of filenames (variable length c-strings) [
	//     - c-string | filename (including path and '/' as dir seperator)
	//   - ]
	//   - 256 bytes | filename
	//   - 8bytes | file size
	//   - fixed chunk size of 4kb
	//   - array of chunk hashes (ids) [
	//     - SHAX bytes
	//   - ]
	HASH_SHA1_INFO3,
	HASH_SHA2_INFO, // hm?

	// id: hash of the content
	// TODO: fixed chunk size or variable (defined in info)
	// if "variable" sized, it can be aliased with TORRENT_V1_CHUNK in the implementation
	HASH_SHA1_CHUNK,
	HASH_SHA2_CHUNK,

	// TODO: design the same thing again for tox? (msg_pack instead of bencode?)
	// id: infohash
	TORRENT_V1_METAINFO,
	// id: sha1
	TORRENT_V1_PIECE, // alias with SHA1_CHUNK?

	// TODO: fix all the v2 stuff here
	// id: infohash
	// in v2, metainfo contains only the root hashes of the merkletree(s)
	TORRENT_V2_METAINFO,
	// id: root hash
	// contains all the leaf hashes for a file root hash
	TORRENT_V2_FILE_HASHES,
	// id: sha256
	// always of size 16KiB, except if last piece in file
	TORRENT_V2_PIECE,
};

// TODO: events
//typedef void NGC_FT1_recv_request_cb(
	//Tox *tox,
	//uint32_t group_number, uint32_t peer_number,
	//const uint8_t* file_id, size_t file_id_size,
	//void* user_data
//);

// return true to accept, false to deny
//typedef bool NGC_FT1_recv_init_cb(
	//Tox *tox,
	//uint32_t group_number, uint32_t peer_number,
	//const uint8_t* file_id, size_t file_id_size,
	//const uint8_t transfer_id,
	//const size_t file_size,
	//void* user_data
//);

//typedef void NGC_FT1_recv_data_cb(
	//Tox *tox,

	//uint32_t group_number,
	//uint32_t peer_number,
	//uint8_t transfer_id,

	//size_t data_offset, const uint8_t* data, size_t data_size,
	//void* user_data
//);

// request to fill data_size bytes into data
//typedef void NGC_FT1_send_data_cb(
	//Tox *tox,

	//uint32_t group_number,
	//uint32_t peer_number,
	//uint8_t transfer_id,

	//size_t data_offset, uint8_t* data, size_t data_size,
	//void* user_data
//);


class NGCFT1 : public ToxEventI, public NGCEXTEventI {
	ToxI& _t;
	ToxEventProviderI& _tep;
	NGCEXTEventProviderI& _neep;

	// TODO: config
	size_t acks_per_packet {3u}; // 3
	float init_retry_timeout_after {5.f}; // 10sec
	float sending_give_up_after {30.f}; // 30sec


	struct Group {
		struct Peer {
			LEDBAT cca{500-4}; // TODO: replace with tox_group_max_custom_lossy_packet_length()-4

			struct RecvTransfer {
				uint32_t file_kind;
				std::vector<uint8_t> file_id;

				enum class State {
					INITED, //init acked, but no data received yet (might be dropped)
					RECV, // receiving data
				} state;

				// float time_since_last_activity ?
				size_t file_size {0};
				size_t file_size_current {0};

				// sequence id based reassembly
				RecvSequenceBuffer rsb;
			};
			std::array<std::optional<RecvTransfer>, 256> recv_transfers;
			size_t next_recv_transfer_idx {0}; // next id will be 0

			struct SendTransfer {
				uint32_t file_kind;
				std::vector<uint8_t> file_id;

				enum class State {
					INIT_SENT, // keep this state until ack or deny or giveup

					SENDING, // we got the ack and are now sending data

					FINISHING, // we sent all data but acks still outstanding????

					// delete
				} state;

				size_t inits_sent {1}; // is sent when creating

				float time_since_activity {0.f};
				size_t file_size {0};
				size_t file_size_current {0};

				// sequence array
				// list of sent but not acked seq_ids
				SendSequenceBuffer ssb;
			};
			std::array<std::optional<SendTransfer>, 256> send_transfers;
			size_t next_send_transfer_idx {0}; // next id will be 0
		};
		std::map<uint32_t, Peer> peers;
	};
	std::map<uint32_t, Group> groups;

	protected:
		bool sendPKG_FT1_REQUEST(uint32_t group_number, uint32_t peer_number, uint32_t file_kind, const uint8_t* file_id, size_t file_id_size);
		bool sendPKG_FT1_INIT(uint32_t group_number, uint32_t peer_number, uint32_t file_kind, uint64_t file_size, uint8_t transfer_id, const uint8_t* file_id, size_t file_id_size);
		bool sendPKG_FT1_INIT_ACK(uint32_t group_number, uint32_t peer_number, uint8_t transfer_id);
		bool sendPKG_FT1_DATA(uint32_t group_number, uint32_t peer_number, uint8_t transfer_id, uint16_t sequence_id, const uint8_t* data, size_t data_size);
		bool sendPKG_FT1_DATA_ACK(uint32_t group_number, uint32_t peer_number, uint8_t transfer_id, const uint16_t* seq_ids, size_t seq_ids_size);

		void updateSendTransfer(float time_delta, uint32_t group_number, uint32_t peer_number, Group::Peer& peer, size_t idx, std::set<LEDBAT::SeqIDType>& timeouts_set);
		void iteratePeer(float time_delta, uint32_t group_number, uint32_t peer_number, Group::Peer& peer);

	public:
		NGCFT1(
			ToxI& t,
			ToxEventProviderI& tep,
			NGCEXTEventProviderI& neep
		);

		void iterate(float delta);

	public: // ft1 api
		// TODO: public variant?
		void NGC_FT1_send_request_private(
			uint32_t group_number, uint32_t peer_number,
			uint32_t file_kind,
			const uint8_t* file_id, size_t file_id_size
		);

		// public does not make sense here
		bool NGC_FT1_send_init_private(
			uint32_t group_number, uint32_t peer_number,
			uint32_t file_kind,
			const uint8_t* file_id, size_t file_id_size,
			size_t file_size,
			uint8_t* transfer_id
		);

	protected:
		bool onEvent(const Events::NGCEXT_ft1_request&) override;
		bool onEvent(const Events::NGCEXT_ft1_init&) override;
		bool onEvent(const Events::NGCEXT_ft1_init_ack&) override;
		bool onEvent(const Events::NGCEXT_ft1_data&) override;
		bool onEvent(const Events::NGCEXT_ft1_data_ack&) override;

	protected:
		//bool onToxEvent(const Tox_Event_Group_Custom_Packet* e) override;
		//bool onToxEvent(const Tox_Event_Group_Custom_Private_Packet* e) override;
};

